﻿using FluentAssertions;
using Leap_Year_kata.BussinessLogic.Services;
using Leap_Year_kata.Tests.Exceptions;

namespace Leap_Year_kata.Tests.Services;

public class LeapYearServiceTests
{
    [Theory]
    [InlineData(2000)]
    [InlineData(4000)]
    public void IsLeapYear_YearDivisibleBy400_ReturnsLeapYear(int leapYear)
    {
        var leapYearService = new LeapYearService();

        new WhenYear(leapYearService)
            .IsGiven(leapYear)
            .IsDivisibleBy400()
            .Then()
            .LeapYearIsTrue();
    }

    [Theory]
    [InlineData(1700)]
    [InlineData(1800)]
    [InlineData(1900)]
    public void IsLeapYear_YearDivisibleBy100ButNotBy400_ReturnsNotLeapYear(int notLeapYear)
    {
        var leapYearService = new LeapYearService();

        new WhenYear(leapYearService)
            .IsGiven(notLeapYear)
            .IsDivisibleBy100()
            .IsNotDivisibleBy400()
            .Then()
            .LeapYearIsFalse();
    }

    [Theory]
    [InlineData(2008)]
    [InlineData(2012)]
    [InlineData(2016)]
    public void IsLeapYear_YearDivisibleBy4ButNotBy100_ReturnsLeapYear(int leapYear)
    {
        var leapYearService = new LeapYearService();

        new WhenYear(leapYearService)
            .IsGiven(leapYear)
            .IsDivisibleBy4()
            .IsNotDivisibleBy100()
            .Then()
            .LeapYearIsTrue();
    }

    [Theory]
    [InlineData(2017)]
    [InlineData(2018)]
    [InlineData(2019)]
    public void IsLeapYear_YearNotDivisibleBy4_ReturnsNotLeapYear(int notLeapYear)
    {
        var leapYearService = new LeapYearService();

        new WhenYear(leapYearService)
            .IsGiven(notLeapYear)
            .IsNotDivisibleBy4()
            .Then()
            .LeapYearIsFalse();
    }

    private sealed class WhenYear
    {

        public WhenYear(LeapYearService leapYearService)
        {
            this.leapYearService = leapYearService;
        }

        private bool _result;
        private int _year;
        private readonly LeapYearService leapYearService;

        public WhenYear IsGiven(int year)
        {
            _year = year;

            return this;
        }

        //theese preparing methods are usually for mocking, but this is only for presenting a priciple
        public WhenYear IsDivisibleBy400()
        {
            if (_year % 400 != 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear IsNotDivisibleBy400()
        {
            if (_year % 400 == 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear IsDivisibleBy100()
        {
            if (_year % 100 != 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear IsNotDivisibleBy100()
        {
            if (_year % 100 == 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear IsDivisibleBy4()
        {
            if (_year % 4 != 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear IsNotDivisibleBy4()
        {
            if (_year % 4 == 0)
                throw new WrongTestArgumentException();

            return this;
        }

        public WhenYear Then()
        {
            _result = leapYearService.IsLeapYear(_year);

            return this;
        }

        public void LeapYearIsTrue() => _result.Should().BeTrue();
        public void LeapYearIsFalse() => _result.Should().BeFalse();
    }
}
